#!/usr/bin/env python
# $Id$

"""
Implements the grid search based on grid.py in libsvm.
Only one function is exported: gridsearch().

This is based on the grid search script of libSVM.
"""

import sys, traceback
from threading import Thread
import queue

from .xvalidate import xvalidate
import mlpy

__all__ = [ "gridsearch" ]

def range_f(begin,end,step):
    "like range, but works on non-integer too"
    seq = []
    while True:
        if step > 0 and begin > end: break
        if step < 0 and begin < end: break
        seq.append(begin)
        begin = begin + step
    return seq

def permute_sequence(seq):
   "Auxiliary function for :func:`calculate_jobs`."
   n = len(seq)
   if n <= 1: return seq

   mid = int(n/2)
   left = permute_sequence(seq[:mid])
   right = permute_sequence(seq[mid+1:])

   ret = [seq[mid]]
   while left or right:
        if left: ret.append(left.pop(0))
        if right: ret.append(right.pop(0))

   return ret

def calculate_jobs(crange,grange):
   """Return a list of lists, containing all the
   (c,g) pairs to be checked during the grid search.
   """
   (c_begin,c_end,c_step) = crange
   (g_begin,g_end,g_step) = grange
   c_seq = permute_sequence(range_f(c_begin,c_end,c_step))
   g_seq = permute_sequence(range_f(g_begin,g_end,g_step))
   nr_c = float(len(c_seq))
   nr_g = float(len(g_seq))
   i = 0
   j = 0
   jobs = []

   while i < nr_c or j < nr_g:
      if i/nr_c < j/nr_g:
        # increase C resolution
        line = []
        for k in range(0,j):
            line.append((c_seq[i],g_seq[k]))
        i = i + 1
        jobs.append(line)
      else:
        # increase g resolution
        line = []
        for k in range(0,i):
            line.append((c_seq[k],g_seq[j]))
        j = j + 1
        jobs.append(line)
   return jobs

class WorkerStopToken:  
   "used to notify the worker to stop"
   pass

class Worker(Thread):
   """One thread for the execution of the grid search.

   The search is parallellised with one Worker per parallel thread.
   """

   def __init__(self,name,job_queue,result_queue,x,y,fold,**kw):
        Thread.__init__(self)
        self.name = name
        self.job_queue = job_queue
        self.result_queue = result_queue
	self.x = x
	self.y = y
	self.fold = fold
	self.xargs = kw
   def run(self):
     while True:
         (cexp,gexp) = self.job_queue.get()
         if cexp is WorkerStopToken:
             self.job_queue.put((cexp,gexp))
             # print 'worker %s stop.' % self.name
             break
         rate = xvalidate(mlpy.LibSvm,self.fold,self.x,self.y,
                  gamma=2.0**gexp,C=2.0**cexp,**self.xargs)
         self.result_queue.put((self.name,cexp,gexp,rate))

def gridsearch(x,y,fold,
      crange=( -5,  15,  2),
      grange=(  3, -15, -2),
      nr_local_worker = 1,
      gnuplot=None,verbosity=0,**kw):
    """
    Perform a grid search based on the given test set and options.
    If gnuplot is given, it should implement the interface of GnuPlot
    object, and will be used to plot and update a contour plot of
    parameter choices during the search.

    The return values is (db,(best_c1, best_g1, best_rate)),
    where db is a list of all the tested parameters with performance,
    best_c1 and best_g1 are the optimal parameters (log c and log gamma),
    and best_rate is the cross-validation accuracy at the optimised
    parameters.
    """

    # put jobs in queue
    jobs = calculate_jobs(crange,grange)
    job_queue = queue.Queue(0)
    result_queue = queue.Queue(0)
    for line in jobs:
      for j in line:
        job_queue.put(j)
    job_queue._put = job_queue.queue.appendleft

    # fire local workers
    for i in range(nr_local_worker):
        Worker('local',job_queue,result_queue,x,y,
	       fold=fold,**kw).start()

    # gather results
    done_jobs = {}

    db = []
    best_rate = -1
    best_c1,best_g1 = None,None

    for line in jobs:
      for (c,g) in line:
        while (c, g) not in done_jobs:
          (worker,c1,g1,rate) = result_queue.get()
          done_jobs[(c1,g1)] = rate
          if (rate > best_rate) or (rate==best_rate and g1==best_g1 and c1<best_c1):
             best_rate = rate
             best_c1,best_g1=c1,g1
             best_c = 2.0**c1
             best_g = 2.0**g1
	  if verbosity > 2:
            print("[%s] %s %s %s (best c=%s, g=%s, rate=%s)" % \
                (worker,c1,g1,rate, best_c, best_g, best_rate))
        db.append((c,g,done_jobs[(c,g)]))
      if gnuplot != None:
         gnuplot.redraw(db,[best_c1, best_g1, best_rate])
    job_queue.put((WorkerStopToken,None))
    return (db,(best_c1, best_g1),(best_c, best_g), best_rate)

